import contextlib
import logging

logger = logging.getLogger("smalld")

redactions = []


def redact_from_logging(to_redact):
    redactions.append(to_redact)


def redact_from(s):
    redacted = str(s)

    for redaction in redactions:
        if redaction in redacted:
            redacted = redacted.replace(redaction, "[[ REDACTED ]]")

    return redacted


class RedactingFilter(logging.Filter):
    def filter(self, record):
        record.msg = redact_from(record.msg)

        if isinstance(record.args, dict):
            record.args = {k: redact_from(v) for k, v in record.args.items()}
        else:
            record.args = tuple(redact_from(arg) for arg in record.args)

        return True


logger.addFilter(RedactingFilter())


@contextlib.contextmanager
def suppress_logging(name):
    suppress = logging.getLogger(name)
    suppress.disabled = True
    yield
    suppress.disabled = False



from collections import abc


def wrap_value(value):
    if isinstance(value, dict):
        return JsonObject(value)
    if isinstance(value, list):
        return JsonArray(value)
    return value


class JsonObject(abc.Mapping):
    __slots__ = ("__data",)

    def __init__(self, data):
        self.__data = data

    def __getitem__(self, key):
        return wrap_value(self.__data[key])

    def __iter__(self):
        return iter(self.__data)

    def __len__(self):
        return len(self.__data)

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{attr}'"
            ) from None

    def __repr__(self):
        return f"<JsonObject {self.__data}>"


class JsonArray(abc.Sequence):
    __slots__ = ("__data",)

    def __init__(self, data):
        self.__data = data

    def __getitem__(self, index):
        return wrap_value(self.__data[index])

    def __len__(self):
        return len(self.__data)

    def __repr__(self):
        return f"<JsonArray {self.__data}>"



import json

from websocket import ABNF, WebSocket, WebSocketException

from .exceptions import NetworkError
from .json_elements import JsonObject
from .logger import logger, suppress_logging
from .ratelimit import GatewayRateLimiter

WebSocketError = (WebSocketException, OSError)


class CloseReason:
    def __init__(self, code=None, reason=""):
        self.code = code
        self.reason = reason

    def __str__(self):
        return f"{self.code}: {self.reason}"

    @staticmethod
    def parse(data):
        code = int.from_bytes(data[:2], "big")
        reason = data[2:].decode("utf-8")
        return CloseReason(code, reason)

    @staticmethod
    def exception(e):
        return CloseReason(reason=f"{type(e).__name__}: {e}")


class Gateway:
    def __init__(self, url):
        self.url = url
        self.ws = WebSocket()
        self.close_reason = None
        self.limiter = GatewayRateLimiter()

    def __iter__(self):
        try:
            self.ws.connect(self.url)
        except WebSocketError as e:
            logger.debug("Exception connecting to gateway.", exc_info=True)
            self.close_reason = CloseReason.exception(e)

        while self.ws.connected:
            try:
                with self.ws.readlock:
                    opcode, data = self.ws.recv_data()
            except WebSocketError as e:
                logger.debug("Exception receiving gateway data.", exc_info=True)
                self.close_reason = CloseReason.exception(e)
                break

            if data and opcode == ABNF.OPCODE_CLOSE:
                self.close_reason = CloseReason.parse(data)
                break

            if data and opcode == ABNF.OPCODE_TEXT:
                decoded_data = data.decode("utf-8")
                logger.debug("Gateway payload received: %s", decoded_data)
                yield JsonObject(json.loads(decoded_data))

        logger.info("Gateway Closed: %s", self.close_reason)

    def send(self, data):
        self.limiter.on_send()
        payload = json.dumps(data)
        logger.debug("Gateway payload sent: %s", payload)
        try:
            self.ws.send(payload)
        except WebSocketError:
            logger.debug("Error sending payload.", exc_info=True)
            raise NetworkError

    def close(self, status=1000):
        # An error is output be websocket-client for non-1000 statuses.
        # During normal operation we use non-1000 statuses and don't want an error logged.
        with suppress_logging("websocket"):
            self.ws.close(status=status)



import os

from setuptools import setup


def readme():
    readme_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")

    with open(readme_path, encoding="utf-8") as f:
        return f.read()


setup(
    name="smalld",
    description="A minimalist python client for the Discord API",
    long_description=readme(),
    long_description_content_type="text/markdown",
    url="https://github.com/princesslana/smalld.py",
    author="Princess Lana",
    author_email="ianagbip1oti@gmail.com",
    license="MIT",
    packages=["smalld", "smalld.resources"],
    package_data={"smalld.resources": ["*"]},
    use_scm_version=True,
    install_requires=["requests>=2.23.0", "websocket_client>=0.57.0"],
    setup_requires=["setuptools-scm==3.3.3"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
